/* Copyright (c) 2003 Theo van Klaveren, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ai.c,v 1.25 2005/04/06 10:20:47 klaveren Exp $
 */
#include <stdio.h>
#include <string.h>

#include <glib.h>

#include "boolball.h"
#include "boolballmodel.h"
#include "ai.h"
#include "move.h"

struct _Ai {
	AiLevel		 level;
};

typedef struct _Range {
	int left;
	int right;
	int range;
} Range;

static int	 ai_easy			(Ai *ai, Game *g, int ball_x);
static int	 ai_normal			(Ai *ai, Game *g, int ball_x);
static int	 ai_hard			(Ai *ai, Game *g, int ball_x);

static int	 ai_simple_score_for_ball	(BoolballModel *model,
						 Player *player,
						 int ball_x);
static int	 ai_score_for_ball		(BoolballModel *model,
						 Player *player,
						 int ball_x);
static Range	*ai_range_for_ball		(BoolballModel *model,
						 Player *player,
						 int ball_x);

Ai *
ai_new(AiLevel level)
{
	Ai *ai;

	ai = g_new0(Ai, 1);
	ai->level = level;

	return ai;
}

void
ai_destroy(Ai *ai)
{
	g_free(ai);
}

void
ai_set_level(Ai *ai, AiLevel level)
{
	ai->level = level;
}

AiLevel
ai_get_level(Ai *ai)
{
	return ai->level;
}

int
ai_get_move(Ai *ai, Game *g)
{
	int *scores;
	int width, height;
	int count, max, row;

	board_model_get_dimensions(BOARD_MODEL(g->model), &width, &height);
	scores = g_malloc0(width * sizeof(int));

	/* Calculate scores */
	for (count = 0; count < width; count++) {
		if (ai->level == AI_EASY)
			scores[count] = ai_easy(ai, g, count);
		else if (ai->level == AI_MEDIUM)
			scores[count] = ai_normal(ai, g, count);
		else if (ai->level == AI_DIFFICULT)
			scores[count] = ai_hard(ai, g, count);
		else
			scores[count] = -1;
	}

	/* Select maximum */
#ifdef DEBUG
	g_print("ai outcome:\n-----------\n");
#endif
	max = -1; row = -1;
	for (count = 0; count < width; count++) {
		if (scores[count] > max) {
			max = scores[count];
			row = count;
		}
#ifdef DEBUG
		if (count % 10 == 0)
			g_print("\n(%2d) ", count);
		printf("%4d ", scores[count]);
#endif
	}
#ifdef DEBUG
	printf ("\n");
#endif

	g_free(scores);

	if (row != -1) {
		return row;
	} else {
		for (count = 0; count < width; count++) {
			if (player_can_drop_ball(g->player, g->model, count)) {
				return count;
			}
		}
	}

	return -1;
}

static int
ai_easy(Ai *ai, Game *g, int ball_x)
{
	BoolballModel *copy = boolball_model_copy(g->model);
	int score;
	
	score = ai_simple_score_for_ball(copy, g->player, ball_x);
	g_object_unref(G_OBJECT(copy));
	return score;
}

static int
ai_normal(Ai *ai, Game *g, int ball_x)
{
	BoolballModel *copy = boolball_model_copy(g->model);
	Player *other_player;
	int my_score, other_score;
	
	my_score = ai_score_for_ball(copy, g->player, ball_x);
	if (my_score == -1)
		return my_score;

	/* Other player has no ball on this spot -> neutral, just return */
	if (g->player->color == PC_RED)
		other_player = g->player_blue;
	else
		other_player = g->player_red;

	if (other_player->balls[ball_x] == CELL_EMPTY) {
		g_object_unref(copy);
		return my_score;
	}

	/* If this spot is now blocked -> very good, calculate damage */
	if (board_model_cell_get(BOARD_MODEL(copy), ball_x, 2) != CELL_EMPTY) {
		BoolballModel *another_copy;
#ifdef DEBUG
		printf("ai_normal: (%d) spot is now blocked\n", ball_x);
#endif
		another_copy = boolball_model_copy(g->model);
		other_score = ai_score_for_ball(another_copy,
						other_player,
						ball_x);
		g_object_unref(another_copy);
#ifdef DEBUG
		printf("ai_normal: damage is %d\n", other_score);
#endif
		/* If this gives us an advantage in points, score higher! */
		if (other_score > 10)
			my_score += (other_score - 10);

		/* We want to keep an advantage of at least 2 balls blocked */
		{
			int my_blocked, other_blocked;

			my_blocked = player_count_blocked_balls(other_player,
								g->model);
			other_blocked = player_count_blocked_balls(g->player,
								   g->model);
			if (my_blocked < other_blocked)
				my_score += 50;
			if (my_blocked <= other_blocked)
				my_score += 25;
			if (my_blocked <= other_blocked + 1)
				my_score += 25;
#ifdef DEBUG
			g_print("ai_normal: my_blocked=%d, other_blocked=%d\n",
					my_blocked, other_blocked);
#endif
		}
	}

	/* Calculate new score for this spot */
	other_score = ai_score_for_ball(copy, other_player, ball_x);

	if (other_score > (my_score + 10)) {
		/* New score higher than old -> bad */
#ifdef DEBUG
		printf("ai: (%d) this will help the player...\n", ball_x);
#endif
		my_score -= ((my_score - other_score) / 2);
	} else if (other_score < (my_score - 10)) {
		/* New score is lower then old -> good */
#ifdef DEBUG
		printf("ai: (%d) this will hurt the player...\n", ball_x);
#endif
		my_score += ((my_score - other_score) * 1.2);
	}

	g_object_unref(copy);
	return my_score;
}

static int
ai_hard(Ai *ai, Game *g, int ball_x)
{
	BoolballModel *copy, *backup;
	Range *range;
	Player *other_player;
	int normal_score;
	int range_score, new_range_score;
	int y;

	/* Other player */
	if (g->player->color == PC_RED)
		other_player = g->player_blue;
	else
		other_player = g->player_red;

	/* Calculate normal score */
	normal_score = ai_normal(ai, g, ball_x);
	if (normal_score == -1)
		return normal_score;

	/* Calculate range of effect */
	copy = boolball_model_copy(g->model);
	range = ai_range_for_ball(copy, g->player, ball_x);
	g_object_unref(copy);

	/*
	 * If the ball fell straight down it's not really interesting for us,
	 * because in some situations this is not likely to change.
	 * This does not apply if we f.e. land (indirectly) on a gray block!
	 */
	if (range->range == 1) {
		gboolean interesting = TRUE;
		gboolean hit_ball = FALSE;
		gint width, height;
		
		board_model_get_dimensions(BOARD_MODEL(g->model), &width, &height);

		for (y = 0; y < height; y++) {
			gint c = board_model_cell_get(BOARD_MODEL(g->model), ball_x, y);
			if (c == CELL_GRAY) {
				/* Interesting! */
				break;
			}
			if (c == CELL_BALL_RED ||
			    c == CELL_BALL_BLUE) {
				/* Might be interesting, if we encounter
				 * a gray block right after the ball */
				hit_ball = TRUE;
				continue;
			}
			if (c == CELL_EMPTY ||
			    c == CELL_BONUS)
				continue;
			if (c == CELL_TELEPORTER) {
				/* Teleporters are uninteresting. */
				interesting = FALSE;
				break;
			}
			/* We encountered balls lying around, but no gray
			 * block (e.g. an arrow). This is
			 * uninteresting. */
			if (hit_ball) {
				interesting = FALSE;
				break;
			}
		}
		/* We hit the bottom without moving left or right,
		 * that's also uninteresting */
		if (y == height)
			interesting = FALSE;

		if (!interesting) {
#ifdef DEBUG
			g_print("ai_hard(%d): Ball is uninteresting\n", ball_x);
#endif
			return 0;
		}
#ifdef DEBUG
		g_print("ai_hard(%d): Ball is interesting\n", ball_x);
#endif
	}

	/*
	 * If the opponent has no balls within range of this one, consider
	 * it uninteresting to drop.
	 */
	{
		gboolean interesting = FALSE;
		for (y = range->left; y <= range->right; y++) {
			if (player_can_drop_ball(other_player, g->model, y))
				interesting = TRUE;
		}
		if (!interesting) {
#ifdef DEBUG
			g_print("ai_hard(%d): Ball is uninteresting\n", ball_x);
#endif
			return 0;
		}
	}
	
	/* Record total score for this range before drop */
	range_score = 0;
	for (y = range->left; y <= range->right; y++) {
		range_score += ai_easy(ai, g, y);
	}

	/* Drop the ball for real */
	copy = boolball_model_copy(g->model);
	ai_simple_score_for_ball(copy, g->player, ball_x);

	/* Change the field in the game to our new one */
	backup = g->model;
	g->model = copy;

	/* Look if we have influenced anything */
	new_range_score = 0;
	for (y = range->left; y <= range->right; y++) {
		new_range_score += ai_easy(ai, g, y);
	}

	/* Restore the real field. */
	g->model = backup;
	g_object_unref(copy);

#ifdef DEBUG
	g_print("ai_hard(%d): adding %d to score\n", ball_x,
			range_score - new_range_score);
#endif
	normal_score += (range_score - new_range_score);
	g_free(range);
	return normal_score;
}

/*
 * Calculates the potential score for the given ball. Returns the amount of
 * points that would be scored, or -1 if the ball can't be dropped. The
 * field is modified.
 */
static int
ai_simple_score_for_ball(BoolballModel *model, Player *player, int ball_x)
{
	Movestate *ms;
	int my_score;
	gboolean falling;

	/* Check whether the ball can actually be dropped. */
	if (!player_can_drop_ball(player, model, ball_x))
		return -1;

	/* Drop the ball in the field. */
	ms = movestate_new();
	ms->fall_x = 2;
	ms->fall_y = ball_x;
	ms->score = 0;

	falling = TRUE;
	while (falling)
		falling = movestate_move(ms, model);

	/* Place the ball where it stopped, and record the score */
	if (!ms->hit_bottom)
		board_model_cell_set(BOARD_MODEL(model), ms->fall_y, ms->fall_x, CELL_BALL_BLUE);

	my_score = ms->score;
	movestate_destroy(ms);

	return my_score;
}

/*
 * Calculates the potential score for the given ball.
 * Try to be a bit more intelligent about it:
 *  - Score negatively on teleporters because we're unceratain about their
 *    outcome,
 *  - Continue falling after gray blocks because they are removed at the end.
 * Returns the amount of points that would be scored, or -1 if the
 * ball can't be dropped. The field is modified.
 */
static int
ai_score_for_ball(BoolballModel *model, Player *player, int ball_x)
{
	Movestate *ms;
	int my_score;
	gboolean falling;

	/* Check whether the ball can actually be dropped. */
	if (!player_can_drop_ball(player, model, ball_x))
		return -1;

	/* Drop the ball in the field. */
	ms = movestate_new();
	ms->fall_x = 2;
	ms->fall_y = ball_x;
	ms->score = 0;

	falling = TRUE;
	while (falling) {
		falling = movestate_move(ms, model);

		/* Continue after gray blocks, they're removed eventually. */
		if (ms->hit_gray_block)
			falling = TRUE;

		/* Don't be too sure about teleporters. */
		if (ms->hit_teleporter)
			ms->score = ms->score - 20;
	}

	/* Place the ball where it stopped, and record the score */
	if (!ms->hit_bottom)
		board_model_cell_set(BOARD_MODEL(model), ms->fall_y, ms->fall_x, CELL_BALL_BLUE);

	my_score = ms->score;
	movestate_destroy(ms);

	return my_score;
}

/*
 * Calculates the range for the given ball's fall. The returned range are
 * the Y coordinates of the fall. Teleporters cause the fall to stop.
 */
static Range *
ai_range_for_ball(BoolballModel *model, Player *player, int ball_x)
{
	Movestate *ms;
	Range *range;
	gboolean falling;
	int left, right;

	/* Create range */
	range = g_new0(Range, 1);

	/* Drop the ball in the field. */
	ms = movestate_new();
	ms->fall_x = 0;
	ms->fall_y = ball_x;
	ms->score = 0;
	left = ball_x;
	right = ball_x;

	falling = TRUE;
	while (falling) {
		falling = movestate_move(ms, model);

		/* Continue after gray blocks, they're removed eventually. */
		if (ms->hit_gray_block)
			falling = TRUE;

		/* Don't do teleporters. They mess up the result. */
		if (ms->hit_teleporter)
			falling = FALSE;

		/* Update range */
		if (ms->fall_y < left)
			left = ms->fall_y;
		if (ms->fall_y > right)
			right = ms->fall_y;
	}

	movestate_destroy(ms);

	range->left = left;
	range->right = right;
	range->range = right - left + 1;
	return range;
}

