/* Copyright (c) 2003 Theo van Klaveren, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: boolball.c,v 1.49 2005/04/07 07:02:39 klaveren Exp $
 */
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include <stdio.h>
#include <glib.h>

#ifdef ENABLE_NLS
#include <locale.h>
#endif

#include <gtk/gtk.h>
#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>

#include "boolball.h"
#include "boolballmodel.h"
#include "boolballview.h"
#include "graphics.h"
#include "player.h"
#include "ai.h"
#include "scorebox.h"
#include "prefs.h"

#define MOVE_DELAY 140
#define ANIM_DELAY 75

struct _Boolball {
	/* Current game being played */
	Game		*game;

	/* Widgets */
	GtkWidget	*bb_window;
	GtkWidget	*bb_vbox;
	GtkWidget	*bb_score_hbox;
	GtkWidget	*about;
	GtkWidget	*score_red;
	GtkWidget	*score_blue;

	/* The model and the view (board) */
	BoolballModel	*bb_model;
	BoolballView	*bb_view;

	/* Status bar */
	GtkWidget	*statusbar;
	gint		 statuscontext;

	/* Timeout id */
	guint		 timeout_id;

	/* Prefs stuff */
	Prefs		*prefs;

	/* AI player */
	Ai		*ai;
};

/* Timeout functions, for animations and such */
gboolean	 move_cb(gpointer data);
gboolean	 move_remove_gray_cb(gpointer data);
gboolean	 move_bottom_slide_cb(gpointer data);
gboolean	 move_player_cb(gpointer data);

/* Move utility functions */
void		 move_score(Boolball *bb, int score);
void		 move_end_score(Boolball *bb, int score);
void		 move_end(Boolball *bb);

void		 board_clicked_cb	(GtkWidget *widget,
					 gint cell_x, gint cell_y,
					 gpointer data);
void		 board_resized_cb	(GtkWidget *widget,
					 gint new_width, gint new_height,
					 gpointer data);

void		 about_destroy_cb	(GtkWidget *widget, gpointer data);
void		 window_destroy_cb	(GtkWidget *widget, gpointer data);
gboolean	 window_expose_cb	(GtkWidget *widget, GdkEventExpose *e, gpointer data);

void		 status_msg_turn(Boolball *bb);
void		 status_msg(Boolball *bb, gchar *msg);
void		 game_over(Boolball *bb);

/* Utilities for building the interface */

static void	 main_build_menu(Boolball *bb);

/* Menu callbacks */

void		 menu_about_cb(gpointer data, guint action, GtkWidget *widget);
void		 menu_new_cb(gpointer data, guint action, GtkWidget *widget);
void		 menu_prefs_cb(gpointer data, guint action, GtkWidget *widget);
void		 menu_quit_cb(gpointer data, guint action, GtkWidget *widget);
void		 menu_scores_cb(gpointer data, guint action, GtkWidget *widget);

/* Menu items */

static GtkItemFactoryEntry menu_items[] = {
	{ N_("/_Game"),			NULL,		NULL,		0,	"<Branch>" },
	{ N_("/Game/_New Game"),	"<control>N",	menu_new_cb,	0,	"<StockItem>",	GTK_STOCK_NEW },
	{ N_("/Game/_Scores..."),	NULL,		menu_scores_cb,	0,	"<StockItem>",	GNOME_STOCK_SCORES },
	{ "/Game/sep1",			NULL,		NULL,		0,	"<Separator>" },
	{ N_("/Game/_Quit"),		"<control>Q",	menu_quit_cb,	0,	"<StockItem>",	GTK_STOCK_QUIT },
	{ N_("/_Settings"),		NULL,		NULL,		0,	"<Branch>" },
	{ N_("/Settings/_Preferences"),	"<control>P",	menu_prefs_cb,	0,	"<StockItem>",	GTK_STOCK_PREFERENCES },
	{ N_("/_Help"),			NULL,		NULL,		0,	"<Branch>" },
	/* /Help/_Contents */
	{ N_("/Help/_About"),		NULL,		menu_about_cb,	0,	"<StockItem>",	GNOME_STOCK_ABOUT }
};

/* Implementation */

void
boolball_drop_ball(Boolball *bb, int ball_x)
{
	Game *game = bb->game;

	status_msg(bb, _("Dropping ball..."));
	movestate_destroy(game->state);
	game->state = movestate_new();
	game->state->fall_x = 2;
	game->state->fall_y = ball_x;
	game->moving = FALSE;

	/* Move the player to the ball. */
	if (bb->timeout_id > 0)
		g_source_remove(bb->timeout_id);
	bb->timeout_id = g_timeout_add(ANIM_DELAY,
				       (GSourceFunc)move_player_cb,
				       bb);
}

void
boolball_new_game(Boolball *bb)
{
	if (bb->timeout_id > 0)
		g_source_remove(bb->timeout_id);

	/* New game and field */
	if (bb->game)
		game_destroy(bb->game);
	boolball_model_set_size(bb->bb_model, prefs_get_field_size(bb->prefs));
	boolball_model_clear(bb->bb_model);
	boolball_model_fill(bb->bb_model);
	bb->game = game_new(bb->bb_model);

	/* New player */
	if (bb->ai)
		ai_destroy(bb->ai);
	bb->ai = ai_new(prefs_get_ai_level(bb->prefs));
	
	move_end(bb);
}

gchar *
i18n_translate(const gchar *path, gpointer data)
{
#ifdef ENABLE_NLS
	return gettext(path);
#else
	return path;
#endif
}

gboolean
move_player_cb(gpointer data)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;
	int ball_x = game->state->fall_y;

	/* Check if we reached our destination */
	if (game->player_pos == ball_x) {

		/* Drop the ball */
		game->falling = TRUE;
		game->player->balls[ball_x] = CELL_EMPTY;

		board_model_freeze(BOARD_MODEL(bb->bb_model));
		board_model_cell_set(BOARD_MODEL(bb->bb_model), ball_x, 1,
				     CELL_EMPTY);
		board_model_cell_set(BOARD_MODEL(bb->bb_model), ball_x, 2,
			  player_get_ball_color(game->player));
		board_model_cell_set(BOARD_MODEL(game->model),
					game->player_pos, 0,
					game->player->color == PC_BLUE ?
						CELL_PLAYER_BLUE_DROP :
						CELL_PLAYER_RED_DROP);
		board_model_thaw(BOARD_MODEL(bb->bb_model));

		/* Move */
		bb->timeout_id = g_timeout_add(MOVE_DELAY,
					       (GSourceFunc)move_cb, bb);
		return FALSE;
	}

	/* Otherwise, move the player */
	board_model_freeze(BOARD_MODEL(bb->bb_model));
	board_model_cell_set(BOARD_MODEL(bb->bb_model), game->player_pos, 0,
				CELL_EMPTY);
	if (game->player_pos < ball_x)
		game->player_pos++;
	if (game->player_pos > ball_x)
		game->player_pos--;
	boolball_model_cell_set_player(BOOLBALL_MODEL(bb->bb_model), game->player_pos,
				game->player->color);
	board_model_thaw(BOARD_MODEL(bb->bb_model));

	return TRUE;
}

gboolean
move_remove_gray_cb(gpointer data)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;
	int scan_x, scan_y;
	int width, height;
	
	board_model_get_dimensions(BOARD_MODEL(bb->bb_model), &width, &height);

	for (scan_y = 0; scan_y < width; scan_y++) {
		for (scan_x = 0; scan_x < height; scan_x++) {
			if (board_model_cell_get(BOARD_MODEL(game->model),
						scan_y, scan_x)
					== CELL_GRAY) {
				board_model_cell_set(BOARD_MODEL(game->model),
						     scan_y, scan_x,
						     CELL_EMPTY);
				return TRUE;
			}
		}
	}
	status_msg(bb, _("Dropping balls..."));
	bb->timeout_id = g_timeout_add(MOVE_DELAY, (GSourceFunc)move_cb, bb);
	return FALSE;
}

gboolean
move_bottom_slide_cb(gpointer data)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;
	Movestate *ms = game->state;
	BoardModel *model = BOARD_MODEL(bb->bb_model);
	gint width, height;

	board_model_freeze(model);
	board_model_cell_set(model, ms->fall_y, ms->fall_x, CELL_EMPTY);

	board_model_get_dimensions(model, &width, &height);
	
	if (ms->fall_y < (width / 2))
		ms->fall_y--;
	else
		ms->fall_y++;

	if (ms->fall_y >= 0 && ms->fall_y < width) {
		board_model_cell_set(model, ms->fall_y, ms->fall_x, player_get_ball_color(game->player));
		board_model_thaw(model);
	} else {
		board_model_thaw(model);
		if (game->finishing)
			bb->timeout_id = g_timeout_add(
						MOVE_DELAY,
						(GSourceFunc)move_cb, bb);
		else
			move_end(bb);
		return FALSE;
	}

	return TRUE;
}

gboolean
move_cb(gpointer data)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;
	Movestate *ms = game->state;
	gint c, old_x, old_y;
	BoardModel *model = BOARD_MODEL(bb->bb_model);
	
	int width, height;
	
	board_model_get_dimensions(model, &width, &height);

	/*
	 * If we're finishing the game and we're not dropping a ball,
	 * find the next one.
	 */
	if (!game->falling && game->finishing) {
		int x, y;

		/* Find a falling ball */
		game->falling = boolball_model_find_free_ball(game->model, &x, &y);
		if (!game->falling) {
			game_over(bb);
			return FALSE;
		} else {
			movestate_destroy(ms);
			game->state = movestate_new();
			ms = game->state;
			/* Switch player to the player whose ball is falling */
			if (board_model_cell_get(model, y, x) == CELL_BALL_BLUE)
				game->player = game->player_blue;
			else
				game->player = game->player_red;
			ms->fall_x = x;
			ms->fall_y = y;
			/* Begin scoring at the current depth */
			ms->score = -(ms->fall_x * 2);
		}
	}

	/* We have a falling ball -> locate it */
	c = board_model_cell_get(model, ms->fall_y, ms->fall_x);
	if (c != CELL_BALL_BLUE && c != CELL_BALL_RED) {
#ifdef DEBUG
		g_print(_("FATAL ERROR: Expected ball, got something else"));
#endif
		game->falling = FALSE;
		game->moving = TRUE;
		return FALSE;
	}

	/* Keep the old spot */
	old_x = ms->fall_x;
	old_y = ms->fall_y;

	/* Move it */
	game->falling = movestate_move(ms, game->model);
	
	/* Did we hit the bottom? */
	if (ms->hit_bottom) {
		/* Ball hit the bottom, start sliding anim */
		move_score(bb, ms->score);
		board_model_freeze(model);
		board_model_cell_set(model, old_y, old_x, CELL_EMPTY);
		board_model_cell_set(model, ms->fall_y, ms->fall_x, player_get_ball_color(game->player));
		board_model_thaw(model);
		game->falling = FALSE;
		bb->timeout_id = g_timeout_add(ANIM_DELAY, (GSourceFunc)move_bottom_slide_cb, data);
		return FALSE;
	}

	/* Did we stop falling? Score and end turn. */
	if (!game->falling) {
		if (game->finishing) {
			move_score(bb, ms->score);
			return TRUE;
		} else {
			move_end_score(bb, ms->score);
			return FALSE;
		}
	}

	/* Did we hit a bonus block? Score and tell the player. */
	if (ms->hit_bonus) {
		status_msg(bb, _("Collected 10 points!"));
		move_score(bb, 10);
		ms->score -= 10;
	}

	/* Did we hit a teleporter? Tell the player. */
	if (ms->teleported) {
		status_msg(bb, _("Ball teleported!"));
	}

	/* Clear the spot where it was and repaint */
	if (old_x != ms->fall_x || old_y != ms->fall_y) {
		board_model_freeze(model);
		if (ms->teleported || ms->left_teleporter)
			board_model_cell_set(model, old_y, old_x, CELL_TELEPORTER);
		else
			board_model_cell_set(model, old_y, old_x, CELL_EMPTY);

		/* Ball not off screen? Draw it. */
		if (ms->fall_y >= 0 &&
		    ms->fall_y < width) {
			board_model_cell_set(model, ms->fall_y, ms->fall_x,
						player_get_ball_color(game->player));
			board_model_thaw(model);
		} else {
			/* Switch turns. */
			board_model_thaw(model);
			move_end(bb);
			if (!game->finishing)
				return FALSE;
		}
	}

	return TRUE;
}

void
move_score(Boolball *bb, int score)
{
	Game *game = bb->game;
	gchar *score_text, *player;

	game->player->score += score;

	if (game->player->color == PC_BLUE) {
		scorebox_set_score(SCOREBOX(bb->score_blue), game->player->score);
		player = g_strdup(_("Blue"));
	} else {
		scorebox_set_score(SCOREBOX(bb->score_red), game->player->score);
		player = g_strdup(_("Red"));
	}

	score_text = g_strdup_printf(_("%s player scored %d points!"),
			player, score);
	status_msg(bb, score_text);

	g_free(score_text);
	g_free(player);
}

void
move_end_score(Boolball *bb, int score)
{
	move_score(bb, score);
	move_end(bb);
}

void
move_end(Boolball *bb)
{
	Game *game = bb->game;

	game->falling = FALSE;

	if (game->finishing)
		return;

	board_model_cell_set(BOARD_MODEL(game->model),
				game->player_pos, 0, CELL_EMPTY);
	game_switch_turns(game);
	status_msg_turn(bb);

	/* Check if that player can move any more balls. */
	if (player_can_move(game->player, game->model)) {
		if (game->player->color == PC_RED &&
		    ai_get_level(bb->ai) != AI_NONE) {
			/* AI's turn */
			int ball_to_drop;
			status_msg(bb,
				_("Computer's turn... (thinking)"));
			ball_to_drop = ai_get_move(bb->ai, game);
#ifdef DEBUG
			g_assert(player_can_drop_ball(game->player,
						      game->model,
						      ball_to_drop));
#endif
			boolball_drop_ball(bb, ball_to_drop);
		} else {
			/* Player's turn */
			game->moving = TRUE;
		}
		return;
	}

	/* No more balls, set flag */
	game->player->finished = TRUE;

	if (game->player_blue->finished && game->player_red->finished) {
		/* End of game */
		game->finishing = TRUE;
		game->falling = FALSE;
		game->moving = FALSE;

		status_msg(bb, _("Removing gray blocks..."));

		/* Remove previous timeout, to make sure that we
		 * remove the gray blocks *before* dropping balls */
		if (bb->timeout_id > 0)
			g_source_remove(bb->timeout_id);
		bb->timeout_id = g_timeout_add(ANIM_DELAY,
					(GSourceFunc)move_remove_gray_cb,
					bb);
	} else {
		/* Switch turns */
		move_end(bb);
	}

}

gboolean
board_key_press_cb(GtkWidget *widget,
		   GdkEventKey *event,
		   gpointer data)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;
	BoardModel *model = BOARD_MODEL(bb->bb_model);
	gint width, height;
	
	board_model_get_dimensions(model, &width, &height);

	if (event->type != GDK_KEY_PRESS)
		return FALSE;

	switch (event->keyval) {
	case GDK_Left:
	case GDK_KP_Left:
	case GDK_KP_4:
		board_model_freeze(model);
		board_model_cell_set(model, game->player_pos, 0, CELL_EMPTY);
		if (game->player_pos > 0)
			game->player_pos--;
		boolball_model_cell_set_player(BOOLBALL_MODEL(bb->bb_model), game->player_pos,
					game->player->color);
		board_model_thaw(model);
		return TRUE;

	case GDK_Right:
	case GDK_KP_Right:
	case GDK_KP_6:
		board_model_freeze(model);
		board_model_cell_set(model, game->player_pos, 0, CELL_EMPTY);
		if (game->player_pos < width - 1)
			game->player_pos++;
		boolball_model_cell_set_player(BOOLBALL_MODEL(bb->bb_model), game->player_pos,
					game->player->color);
		board_model_thaw(model);
		return TRUE;

	case GDK_Down:
	case GDK_KP_Down:
	case GDK_KP_2:
		if (player_can_drop_ball(game->player, game->model,
					 game->player_pos)) {
			boolball_drop_ball(bb, game->player_pos);
		}
		return TRUE;

	default:
		return FALSE;
	}
	/* NOTREACHED */
}

void
board_clicked_cb(GtkWidget *widget, gint cell_x, gint cell_y, gpointer data)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;

	if (game->finishing || !game->playing || !game->moving)
		return;

	if (cell_y < 1 || cell_y >=2)
		return;

	/* Is there a ball at that spot? If so, drop it! */
	if (player_can_drop_ball(game->player, game->model, cell_x)) {
		boolball_drop_ball(bb, cell_x);
	}

	return;
}

void
board_resized_cb(GtkWidget *widget, gint new_width, gint new_height, gpointer data)
{
	graphics_load_at_size(new_width, new_height);
}

void
about_destroy_cb(GtkWidget *widget, gpointer data)
{
	Boolball *bb = (Boolball *)data;

	bb->about = NULL;
}

void
window_destroy_cb(GtkWidget *widget, gpointer data)
{
	Boolball *bb = (Boolball *)data;

	if (bb->timeout_id > 0)
		g_source_remove(bb->timeout_id);
	gtk_main_quit();
}

gboolean
window_expose_cb(GtkWidget *widget, GdkEventExpose *e, gpointer data)
{
	Boolball *bb = (Boolball *)data;

	if (bb->prefs != NULL) {
		gint w,h;
		gtk_window_get_size(GTK_WINDOW(bb->bb_window), &w, &h);
		prefs_set_window_size(bb->prefs, w, h);
	}
	return FALSE;
}

void
status_msg_turn(Boolball *bb)
{
	gchar *msg;
	msg = g_strdup_printf(_("%s player: Choose a ball to drop!"),
			bb->game->player->color == PC_RED ?
				_("Red") : _("Blue"));
	status_msg(bb, msg);
	g_free(msg);
}

void
status_msg(Boolball *bb, gchar *msg)
{
	GtkStatusbar *statusbar = GTK_STATUSBAR(bb->statusbar);

	gtk_statusbar_pop(statusbar, bb->statuscontext);
	gtk_statusbar_push(statusbar, bb->statuscontext, msg);
}

void
game_over(Boolball *bb)
{
	Game *game = bb->game;
	GtkWidget *dialog;
	gchar *msg;
	int result;
	int blue_score, red_score;
	const gchar *level;

	game->playing = FALSE;

	if (game->player_blue->score > game->player_red->score)
		msg = g_strdup(_("Blue player has won!"));
	else if (game->player_red->score > game->player_blue->score)
		msg = g_strdup(_("Red player has won!"));
	else
		msg = g_strdup(_("It's a tie!"));

	status_msg(bb, _("Game over!"));
	dialog = gtk_message_dialog_new(GTK_WINDOW(bb->bb_window), 0,
			GTK_MESSAGE_INFO, GTK_BUTTONS_OK, msg);
	result = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	level = boolball_model_get_size_string(game->model);
	blue_score = gnome_score_log(game->player_blue->score, level, TRUE);
	if (blue_score > 0) {
		dialog = gnome_scores_display_with_pixmap(NULL,
							 "boolball",
							 level,
							 blue_score);
		if (dialog)
			result = gtk_dialog_run(GTK_DIALOG(dialog));
	}

	/* Don't record score for computer players. */
	if (ai_get_level(bb->ai) == AI_NONE) {
		red_score = gnome_score_log(game->player_blue->score,
					    level,
					    TRUE);
		if (red_score > 0) {
			dialog = gnome_scores_display_with_pixmap(NULL,
								 "boolball",
								 level,
								 red_score);
			if (dialog)
				result = gtk_dialog_run(GTK_DIALOG(dialog));
		}
	}
}

void
menu_about_cb(gpointer data, guint action, GtkWidget *widget)
{
	Boolball *bb = (Boolball *)data;

	if (bb->about == NULL) {
		const gchar *authors[] = {
			"Theo van Klaveren <t.vanklaveren@student.utwente.nl>",
			NULL
		};
		const gchar *documenters[] = {
			NULL
		};
		gchar *translator_credits = _("translator_credits");

		bb->about = gnome_about_new(
				_("Boolball"), VERSION,
				"Copyright \xc2\xa9 2003 Theo van Klaveren",
				_("Drop balls as low as possible to score "
				  "points and outsmart your opponent."),
				authors,
				documenters,
				strcmp (translator_credits,
					"translator_credits") != 0 ?
						translator_credits : NULL,
				NULL);
		g_signal_connect(G_OBJECT(bb->about), "destroy",
				 G_CALLBACK(about_destroy_cb), bb);
		gtk_window_set_transient_for(GTK_WINDOW(bb->about),
					     GTK_WINDOW(bb->bb_window));
		gtk_widget_show(bb->about);
	} else {
		gtk_window_present(GTK_WINDOW(bb->about));
	}
}

void
menu_new_cb(gpointer data, guint action, GtkWidget *widget)
{
	Boolball *bb = (Boolball *)data;

	boolball_new_game(bb);
}

void
menu_prefs_cb(gpointer data, guint action, GtkWidget *widget)
{
	Boolball *bb = (Boolball *)data;

	prefs_show(bb->prefs);
}

void
menu_quit_cb(gpointer data, guint action, GtkWidget *widget)
{
	window_destroy_cb(widget, data);
}

void
menu_scores_cb(gpointer data, guint action, GtkWidget *widget)
{
	Boolball *bb = (Boolball *)data;
	Game *game = bb->game;
	GtkWidget *dialog;
	const gchar *level;

	level = boolball_model_get_size_string(game->model);
	dialog = gnome_scores_display_with_pixmap(NULL, "boolball", level, -1);
	if (dialog)
		gtk_dialog_run(GTK_DIALOG(dialog));
}

static void
main_build_menu(Boolball *bb)
{
	GtkItemFactory *factory;
	GtkAccelGroup *group;
	int n_entries = (sizeof(menu_items) / sizeof(menu_items[0]));
	GtkWidget *menubar;

	/* Create the menu */
	group = gtk_accel_group_new();
	factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", group);
	gtk_item_factory_set_translate_func(factory, i18n_translate,
			NULL, NULL);
	gtk_item_factory_create_items(factory, n_entries, menu_items, bb);
	menubar = gtk_item_factory_get_widget(factory, "<main>");

	/* Add it */
	gtk_box_pack_start(GTK_BOX(bb->bb_vbox), menubar, FALSE, TRUE, 0);
	gtk_window_add_accel_group(GTK_WINDOW(bb->bb_window), group);
}

int
main(int argc, char *argv[])
{
	Boolball *bb;
	gint w,h;

	/* Initialize */
	gnome_score_init("boolball");
#ifdef ENABLE_NLS
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
	textdomain(PACKAGE);
#endif
	gnome_program_init("Boolball", VERSION,
			   LIBGNOMEUI_MODULE,
			   argc, argv,
			   GNOME_PARAM_APP_DATADIR, DATADIR,
			   NULL);
	graphics_load_at_size(24, 24); /* FIXME shouldn't be here at all */
	bb = g_new0(Boolball, 1);
	bb->prefs = prefs_new();
	prefs_get_window_size(bb->prefs, &w, &h);

	/* Setup window */
	bb->bb_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_default_size(GTK_WINDOW(bb->bb_window), w, h);
	g_signal_connect(G_OBJECT(bb->bb_window), "destroy",
			 G_CALLBACK(window_destroy_cb), bb);
	g_signal_connect(G_OBJECT(bb->bb_window), "expose_event",
			G_CALLBACK(window_expose_cb), bb);
	bb->bb_vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(bb->bb_window), bb->bb_vbox);
	main_build_menu(bb);

	/* Score boxes */
	bb->bb_score_hbox = gtk_hbox_new(TRUE, 6);
	gtk_box_pack_start(GTK_BOX(bb->bb_vbox),
				bb->bb_score_hbox, FALSE, TRUE, 0);

	bb->score_blue = scorebox_new(PC_BLUE);
	gtk_box_pack_start(GTK_BOX(bb->bb_score_hbox),
			bb->score_blue,
			TRUE, TRUE, 0);
	bb->score_red = scorebox_new(PC_RED);
	gtk_box_pack_start(GTK_BOX(bb->bb_score_hbox),
			bb->score_red,
			TRUE, TRUE, 0);

	/* Status bar */
	bb->statusbar = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(bb->statusbar), FALSE);
	bb->statuscontext = gtk_statusbar_get_context_id(
			GTK_STATUSBAR(bb->statusbar), "Boolball status");
	gtk_box_pack_end(GTK_BOX(bb->bb_vbox), bb->statusbar, FALSE, FALSE, 0);

	/* Setup game board */
	bb->bb_view = boolball_view_new();
	bb->bb_model = boolball_model_new();
	board_view_set_square(BOARD_VIEW(bb->bb_view), TRUE);
	board_view_set_model(BOARD_VIEW(bb->bb_view), BOARD_MODEL(bb->bb_model));
	g_signal_connect(G_OBJECT(bb->bb_view), "board_resized",
			 G_CALLBACK(board_resized_cb), bb);
	g_signal_connect(G_OBJECT(bb->bb_view), "board_clicked",
			 G_CALLBACK(board_clicked_cb), bb);
	g_signal_connect(G_OBJECT(bb->bb_window), "key_press_event",
			 G_CALLBACK(board_key_press_cb), bb);
	gtk_box_pack_start(GTK_BOX(bb->bb_vbox),
			   GTK_WIDGET(bb->bb_view), TRUE, TRUE, 0);

	/* And go!*/
	boolball_new_game(bb);
	gtk_widget_show_all(bb->bb_window);
	gtk_main();
	return 0;
}

