#include <string.h>

#include "boolballmodel.h"
#include "player.h"

typedef struct _ModelSizeData {
	gint height;
	gint width;
	gint num_left_arrow;
	gint num_right_arrow;
	gint num_gray_block;
	gint num_teleporter;
	gint num_bonus;
} ModelSizeData;

#define TELEPORTERS_MAX 10
static const ModelSizeData ms_small_data =  { 14, 20, 30, 30, 12, 3, 8 };
static const ModelSizeData ms_medium_data = { 18, 30, 60, 60, 20, 5, 10 };
static const ModelSizeData ms_large_data =  { 22, 40, 80, 80, 35, 10, 20 };

const gchar *ms_sizestrings[3] = {"small", "medium", "large" };

G_DEFINE_TYPE(BoolballModel, boolball_model, BOARD_MODEL_TYPE);

typedef struct _BoolballModelPrivate {
	/* Size data for this board */
	ModelSizeData	 msd;
	gint		 modelsize;

	/* Teleporter locations */
	gint		 teleporters_x[TELEPORTERS_MAX];
	gint		 teleporters_y[TELEPORTERS_MAX];
} BoolballModelPrivate;

#define BOOLBALL_MODEL_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), BOOLBALL_MODEL_TYPE, BoolballModelPrivate))

static void _place_random(BoolballModel *model, int what, int howmany);
static void _place_random_teleporters(BoolballModel *model, int howmany);
static gboolean _ball_can_drop(BoolballModel *model, int x, int y);

static void
boolball_model_init(BoolballModel *model)
{

}

static void
boolball_model_class_init(BoolballModelClass *klass)
{
	g_type_class_add_private(klass, sizeof(BoolballModelPrivate));
}

BoolballModel *
boolball_model_new()
{
	return BOOLBALL_MODEL(g_object_new(BOOLBALL_MODEL_TYPE, NULL));
}

BoolballModel *
boolball_model_copy(BoolballModel *orig)
{
	BoolballModel *copy;
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(orig);
	BoolballModelPrivate *privcopy;

	copy = boolball_model_new();
	board_model_clone(BOARD_MODEL(orig), BOARD_MODEL(copy));
	privcopy = BOOLBALL_MODEL_GET_PRIVATE(copy);
	g_memmove(privcopy, priv, sizeof(BoolballModelPrivate));
	return copy;
}

void
boolball_model_set_size(BoolballModel *model, gint modelsize)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);
	const ModelSizeData *msdnew;

	if (model == NULL)
		return;
	priv->modelsize = modelsize;

	switch(modelsize) {
		case MODEL_SMALL:
			msdnew = &ms_small_data;
			break;
		case MODEL_LARGE:
			msdnew = &ms_large_data;
			break;
		case MODEL_MEDIUM:
		default:
			msdnew = &ms_medium_data;
			break;
	}
	memcpy(&priv->msd, msdnew, sizeof(ModelSizeData));
	/*
	 * We add 3 to the height of the board here. These are for:
	 *  - The player movement area;
	 *  - The player's balls;
	 *  - The bottom row.
	 */
	board_model_set_dimensions(BOARD_MODEL(model), priv->msd.width,
			priv->msd.height + 3);
}

const gchar *
boolball_model_get_size_string(BoolballModel *model)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);

	return ms_sizestrings[priv->modelsize];
}

void
boolball_model_clear(BoolballModel *model) {
	int i,j;
	int imax, jmax;

	board_model_get_dimensions(BOARD_MODEL(model), &imax, &jmax);

	for (i = 0; i < imax; i++) {
		for (j = 0; j < jmax; j++) {
			board_model_cell_set(BOARD_MODEL(model), i, j,
					CELL_EMPTY);
		}
	}
}

void
boolball_model_fill(BoolballModel *model)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);

	_place_random(model, CELL_ARROW_LEFT, priv->msd.num_left_arrow);
	_place_random(model, CELL_ARROW_RIGHT, priv->msd.num_right_arrow);
	_place_random(model, CELL_GRAY, priv->msd.num_gray_block);
	_place_random(model, CELL_BONUS, priv->msd.num_bonus);
	_place_random_teleporters(model, priv->msd.num_teleporter);
}

void
boolball_model_cell_set_player(BoolballModel *model, gint y, gint color)
{
	gint cell;

	if (color == PC_RED) {
		if (y % 2 == 1)
			cell = CELL_PLAYER_RED_ODD;
		else
			cell = CELL_PLAYER_RED_EVEN;
		/* FIXME */
		/*
		if (b->game->falling)
			pb = graphics_get_for_cell(CELL_PLAYER_BLUE_DROP);
			*/
	} else {
		if (y % 2 == 1)
			cell = CELL_PLAYER_BLUE_ODD;
		else
			cell = CELL_PLAYER_BLUE_EVEN;
		/* FIXME */
		/*
		if (b->game->falling)
			pb = graphics_get_for_cell(CELL_PLAYER_BLUE_DROP);
		*/
	}

	board_model_cell_set(BOARD_MODEL(model), y, 0, cell);
}

gboolean
boolball_model_find_free_ball(BoolballModel *model, int *x, int *y)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);
	int scan_x, scan_y;

	for (scan_y = 0; scan_y < priv->msd.width; scan_y++) {
		for (scan_x = priv->msd.height - 1; scan_x >= 2; scan_x--) {
			gint c = board_model_cell_get(BOARD_MODEL(model),
							 scan_y, scan_x);
			if (c == CELL_BALL_BLUE || c == CELL_BALL_RED) {
				if (_ball_can_drop(model, scan_x, scan_y)) {
					*x = scan_x;
					*y = scan_y;
					return TRUE;
				}
			}
		}
	}
	return FALSE;
}

void
boolball_model_find_teleporter(BoolballModel *model, int *x, int *y)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);

	g_print("finding teleporter\n");
	while(1) {
		gint pick;
		gint tp_x, tp_y;
	
		pick = g_random_int_range(0, priv->msd.num_teleporter);
		tp_x = priv->teleporters_x[pick];
		tp_y = priv->teleporters_y[pick];
		g_print("picking %d: (%d, %d)\n", pick, tp_x, tp_y);
		if (board_model_cell_get(BOARD_MODEL(model), tp_x, tp_y)
			== CELL_TELEPORTER) {
			*x = tp_x;
			*y = tp_y;
			return;
		}
	}

	g_assert_not_reached();
}

static void
_place_random(BoolballModel *model, int what, int howmany)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);
	int count;

	for (count = 0; count < howmany; count++) {
		guchar x,y;

		/* Keep top row free of obstacles, so start at row 1 */
		y = (guchar)g_random_int_range(1, priv->msd.height);
		x = (guchar)g_random_int_range(0, priv->msd.width);
		board_model_cell_set(BOARD_MODEL(model), x, y + 2, what);
	}
}

static void
_place_random_teleporters(BoolballModel *model, int howmany)
{
	BoolballModelPrivate *priv = BOOLBALL_MODEL_GET_PRIVATE(model);
	int count;

	for (count = 0; count < howmany; count++) {
		guchar x,y;

		/*
		 * Keep top row free of obstacles, so start at row 1.
		 * Don't place teleporters on bottom row, so take 1 off max.
		 */
		y = (guchar)g_random_int_range(1, priv->msd.height - 1);
		x = (guchar)g_random_int_range(0, priv->msd.width);

		if (board_model_cell_get(BOARD_MODEL(model), x, y) == CELL_TELEPORTER) {

			/* Don't overwrite existing teleporters, try again */
			count--;
			continue;
		}

		/* Save it's position */
		priv->teleporters_x[count] = x;
		priv->teleporters_y[count] = y + 2;

		board_model_cell_set(BOARD_MODEL(model), x, y + 2, CELL_TELEPORTER);
	}
}

/*
 * Returns whether a given ball is able to fall.
 */
static gboolean
_ball_can_drop(BoolballModel *model, int x, int y)
{
	gint under = board_model_cell_get(BOARD_MODEL(model), y, x + 1);
	if (under == CELL_EMPTY)
		return TRUE;
	else
		return FALSE;
}
