/* Copyright (c) 2003 Theo van Klaveren, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: move.c,v 1.7 2005/04/06 10:20:48 klaveren Exp $
 */
#include <glib.h>

#include "move.h"

static gboolean		is_free_space(gint c);

Movestate *
movestate_new()
{
	Movestate *ms;

	ms = g_new0(Movestate, 1);

	ms->hit_teleporter = FALSE;
	ms->hit_gray_block = FALSE;
	ms->hit_bonus = FALSE;
	ms->hit_bottom = FALSE;
	ms->left_teleporter = FALSE;

	ms->teleported = FALSE;
	ms->roll_left = FALSE;
	ms->roll_right = FALSE;

	ms->fall_x = 0;
	ms->fall_y = 0;
	ms->score = 0;
	return ms;
}

void
movestate_destroy(Movestate *ms)
{
	g_free(ms);
}

gboolean
movestate_move(Movestate *ms, BoolballModel *model)
{
	gint c;
	gint width, height;
	
	board_model_get_dimensions(BOARD_MODEL(model), &width, &height);

	/* Clear state of last turn */
	if (ms->left_teleporter)
		ms->left_teleporter = FALSE;
	if (ms->teleported) {
		ms->teleported = FALSE;
		ms->left_teleporter = TRUE;
	}
	if (ms->hit_bonus)
		ms->hit_bonus = FALSE;

	/* Did we hit a teleporter last turn? */
	if (ms->hit_teleporter) {
		ms->teleported = TRUE;
		ms->hit_teleporter = FALSE;
		boolball_model_find_teleporter(model, &ms->fall_y, &ms->fall_x);
		return TRUE;
	}

	ms->fall_x++;
	if (ms->fall_x >= height) {
		/* Bottom, score 60 points and break. */
		ms->fall_x--;
		ms->hit_bottom = TRUE;
		ms->score += 60;
		return FALSE;
	}

	/* Otherwise, look at what's next. */
	switch(board_model_cell_get(BOARD_MODEL(model), ms->fall_y, ms->fall_x)) {
	case CELL_ARROW_LEFT:
		ms->roll_left = FALSE;
		ms->roll_right = FALSE;
		board_model_cell_set(BOARD_MODEL(model), ms->fall_y, ms->fall_x, CELL_ARROW_RIGHT);
		ms->fall_x--;
		if (ms->fall_y > 0) {
			c = board_model_cell_get(BOARD_MODEL(model), ms->fall_y - 1, ms->fall_x);
		} else {
			ms->score += ((ms->fall_x + 1) * 2);
			return FALSE;
		}
		if (is_free_space(c)) {
			ms->fall_y--;
			if (c != CELL_TELEPORTER)
				ms->roll_left = TRUE;
			if (c == CELL_BONUS) {
				ms->hit_bonus = TRUE;
				ms->score += 10;
			}
			if (c == CELL_TELEPORTER)
				ms->hit_teleporter = TRUE;
		} else {
			ms->score += ((ms->fall_x + 1) * 2);
			return FALSE;
		}
		break;

	case CELL_ARROW_RIGHT:
		ms->roll_left = FALSE;
		ms->roll_right = FALSE;
		board_model_cell_set(BOARD_MODEL(model), ms->fall_y, ms->fall_x, CELL_ARROW_LEFT);
		ms->fall_x--;
		if (ms->fall_y < width - 1) {
			c = board_model_cell_get(BOARD_MODEL(model), ms->fall_y + 1, ms->fall_x);
		} else {
			ms->score += ((ms->fall_x + 1) * 2);
			return FALSE;
		}
		if (is_free_space(c)) {
			ms->fall_y++;
			if (c != CELL_TELEPORTER)
				ms->roll_right = TRUE;
			if (c == CELL_BONUS) {
				ms->hit_bonus = TRUE;
				ms->score += 10;
			}
			if (c == CELL_TELEPORTER)
				ms->hit_teleporter = TRUE;
		} else {
			ms->score += ((ms->fall_x + 1) * 2);
			return FALSE;
		}
		break;

	case CELL_GRAY:
		ms->roll_left = FALSE;
		ms->roll_right = FALSE;

		/*
		 * If we hit a gray block *last* turn and are still called
		 * again, continue instead of stopping. This is probably
		 * used by the AI.
		 */
		if (!ms->hit_gray_block) {
			ms->fall_x--;
			ms->score += ((ms->fall_x + 1) * 2);
			ms->hit_gray_block = TRUE;
			return FALSE;
		}

		ms->score -= (ms->fall_x * 2);
		ms->hit_gray_block = FALSE;
		break;

	case CELL_BALL_BLUE:
	case CELL_BALL_RED:

		/* We're not going to fall in any case below. */
		ms->fall_x--;

		/* Case 1: We're rolling left */
		if (ms->roll_left && ms->fall_y > 0) {
			c = board_model_cell_get(BOARD_MODEL(model),
				      ms->fall_y - 1,
				      ms->fall_x);
			if (is_free_space(c)) {
				ms->fall_y--;
				if (c == CELL_TELEPORTER)
					ms->hit_teleporter = TRUE;
				if (c == CELL_BONUS) {
					ms->hit_bonus = TRUE;
					ms->score += 10;
				}
				break;
			}
		}

		/* Case 2: We're rolling right */
		if (ms->roll_right &&
		    ms->fall_y < width - 1) {
			c = board_model_cell_get(BOARD_MODEL(model),
				      ms->fall_y + 1,
				      ms->fall_x);
			if (is_free_space(c)) {
				ms->fall_y++;
				if (c == CELL_TELEPORTER)
					ms->hit_teleporter = TRUE;
				if (c == CELL_BONUS) {
					ms->hit_bonus = TRUE;
					ms->score += 10;
				}
				break;
			}
		}

		/* Case 3: We're not rolling, or the rolling was blocked */
		ms->score += ((ms->fall_x + 1) * 2);
		return FALSE;

	case CELL_TELEPORTER:
		ms->roll_left = FALSE;
		ms->roll_right = FALSE;
		ms->hit_teleporter = TRUE;
		break;

	case CELL_BONUS:
		ms->roll_left = FALSE;
		ms->roll_right = FALSE;
		ms->hit_bonus = TRUE;
		ms->score += 10;
		break;

	default:
		ms->roll_left = FALSE;
		ms->roll_right = FALSE;
		break;
	}

	return TRUE;
}

static gboolean
is_free_space(gint c)
{
	return (c == CELL_BONUS ||
	        c == CELL_EMPTY ||
		c == CELL_TELEPORTER);
}

