/* Copyright (c) 2003 Theo van Klaveren, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: prefs.c,v 1.6 2005/04/06 20:12:11 klaveren Exp $
 */
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include <glib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>

#include "prefs.h"
#include "ai.h"

struct _Prefs {
	GladeXML	*xml;
	GConfClient	*client;

	gint		 size;
	AiLevel		 level;
};

/* Utilities for the prefs dialog */
static void	 load_xml(Prefs *p);
static void	 prefs_update_ai_level(Prefs *p);
static void	 prefs_update_board_size(Prefs *p);

/* GConf notification */
void		 gconf_ai_level_notify(GConfClient *client,
				       guint cnxn_id,
				       GConfEntry *entry,
				       gpointer user_data);
void		 gconf_board_size_notify(GConfClient *client,
					 guint cnxn_id,
					 GConfEntry *entry,
					 gpointer user_data);

/* Glade callbacks */
static void	 on_prefs_dialog_response(GtkWidget *widget,
					  gint response,
					  gpointer data);
static void	 on_prefs_dialog_destroy(GtkWidget *widget, gpointer data);
static void	 on_rb_small_toggled(GtkWidget *widget, gpointer data);
static void	 on_rb_normal_toggled(GtkWidget *widget, gpointer data);
static void	 on_rb_large_toggled(GtkWidget *widget, gpointer data);
static void	 on_rb_pvp_toggled(GtkWidget *widget, gpointer data);
static void	 on_rb_pveasy_toggled(GtkWidget *widget, gpointer data);
static void	 on_rb_pvmedium_toggled(GtkWidget *widget, gpointer data);
static void	 on_rb_pvdiff_toggled(GtkWidget *widget, gpointer data);

Prefs *
prefs_new()
{
	Prefs *p = g_new0(Prefs, 1);

	/* Setup GConf */
	p->client = gconf_client_get_default();
	gconf_client_add_dir(p->client, GCONF_APP_DIR, 
			GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);

	p->size = gconf_client_get_int(p->client, GCONF_KEY_BOARD_SIZE, NULL);
	if (p->size < MODEL_SMALL || p->size > MODEL_LARGE) {
		gconf_client_set_int(p->client, GCONF_KEY_BOARD_SIZE,
				     MODEL_MEDIUM, NULL);
		p->size = MODEL_MEDIUM;
	}

	p->level = gconf_client_get_int(p->client, GCONF_KEY_AI_LEVEL, NULL);
	if (p->level < AI_EASY || p->level > AI_NONE) {
		gconf_client_set_int(p->client, GCONF_KEY_AI_LEVEL,
				     AI_MEDIUM, NULL);
		p->level = AI_MEDIUM;
	}

	gconf_client_notify_add(p->client, GCONF_KEY_AI_LEVEL,
			gconf_ai_level_notify, p, NULL, NULL);
	gconf_client_notify_add(p->client, GCONF_KEY_BOARD_SIZE,
			gconf_board_size_notify, p, NULL, NULL);

	return p;
}

void
prefs_destroy(Prefs *p)
{
	if (p->xml)
		g_object_unref(G_OBJECT(p->xml));
	g_free(p);
}

void
prefs_show(Prefs *p)
{
	GtkWidget *dialog;

	if (p->xml != NULL) {
		dialog = glade_xml_get_widget(p->xml, "prefs_dialog");
		gtk_window_present(GTK_WINDOW(dialog));
		return;
	}

	load_xml(p);
	prefs_update_ai_level(p);
	prefs_update_board_size(p);
	dialog = glade_xml_get_widget(p->xml, "prefs_dialog");
	gtk_widget_show_all(dialog);
}

gint
prefs_get_field_size(Prefs *p)
{
	return p->size;
}

AiLevel
prefs_get_ai_level(Prefs *p)
{
	return p->level;
}

gboolean
prefs_get_window_size(Prefs *p, gint *w, gint *h)
{
	gint gw, gh;
	gw = gconf_client_get_int(p->client, GCONF_KEY_WINDOW_WIDTH, NULL);
	if (gw < 200)
		gw = 200;
	gh = gconf_client_get_int(p->client, GCONF_KEY_WINDOW_HEIGHT, NULL);
	if (gh < 100)
		gh = 100;
	*w = gw;
	*h = gh;
	return TRUE;
}

void
prefs_set_window_size(Prefs *p, gint w, gint h)
{
	gconf_client_set_int(p->client, GCONF_KEY_WINDOW_WIDTH, w, NULL);
	gconf_client_set_int(p->client, GCONF_KEY_WINDOW_HEIGHT, h, NULL);
}

void
load_xml(Prefs *p)
{
	GladeXML *xml;

	/* Load GLADE XML file */
	p->xml = glade_xml_new(GLADEDIR "/boolball.glade", NULL, PACKAGE);
	xml = p->xml;

	/* Connect signals */
	glade_xml_signal_connect_data(xml, "on_prefs_dialog_response",
			(GCallback)on_prefs_dialog_response, p);
	glade_xml_signal_connect_data(xml, "on_prefs_dialog_destroy",
			(GCallback)on_prefs_dialog_destroy, p);
	glade_xml_signal_connect_data(xml, "on_rb_small_toggled",
			(GCallback)on_rb_small_toggled, p);
	glade_xml_signal_connect_data(xml, "on_rb_normal_toggled",
			(GCallback)on_rb_normal_toggled, p);
	glade_xml_signal_connect_data(xml, "on_rb_large_toggled",
			(GCallback)on_rb_large_toggled, p);
	glade_xml_signal_connect_data(xml, "on_rb_pvp_toggled",
			(GCallback)on_rb_pvp_toggled, p);
	glade_xml_signal_connect_data(xml, "on_rb_pveasy_toggled",
			(GCallback)on_rb_pveasy_toggled, p);
	glade_xml_signal_connect_data(xml, "on_rb_pvmedium_toggled",
			(GCallback)on_rb_pvmedium_toggled, p);
	glade_xml_signal_connect_data(xml, "on_rb_pvdiff_toggled",
			(GCallback)on_rb_pvdiff_toggled, p);
}

static void
prefs_update_ai_level(Prefs *p)
{
	GtkWidget *button;
	AiLevel level = p->level;

	switch(level) {
	case AI_EASY:
		button = glade_xml_get_widget(p->xml, "rb_pveasy");
		break;

	case AI_MEDIUM:
		button = glade_xml_get_widget(p->xml, "rb_pvmedium");
		break;

	case AI_DIFFICULT:
		button = glade_xml_get_widget(p->xml, "rb_pvdiff");
		break;

	case AI_NONE:
		button = glade_xml_get_widget(p->xml, "rb_pvp");
		break;

	default:
		return;
	}

	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
}

static void
prefs_update_board_size(Prefs *p)
{
	GtkWidget *button;
	gint size = p->size;

	switch(size) {
	case MODEL_SMALL:
		button = glade_xml_get_widget(p->xml, "rb_small");
		break;

	case MODEL_MEDIUM:
		button = glade_xml_get_widget(p->xml, "rb_normal");
		break;

	case MODEL_LARGE:
		button = glade_xml_get_widget(p->xml, "rb_large");
		break;

	default:
		return;
	}
	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
}

void
gconf_ai_level_notify(GConfClient *client,
		      guint cnxn_id,
		      GConfEntry *entry,
		      gpointer user_data)
{
	GConfValue *value;
	Prefs *p = (Prefs *)user_data;
	AiLevel level;

	value = gconf_entry_get_value(entry);
	level = gconf_value_get_int(value);
	if (level < AI_EASY || level > AI_NONE)
		level = AI_MEDIUM;
	p->level = level;

	/* Update preferences dialog */
	if (p->xml)
		prefs_update_ai_level(p);
}

void		
gconf_board_size_notify(GConfClient *client,
			guint cnxn_id,
			GConfEntry *entry,
			gpointer user_data)
{
	GConfValue *value;
	Prefs *p = (Prefs *)user_data;
	gint size;

	value = gconf_entry_get_value(entry);
	size = gconf_value_get_int(value);
	if (size < MODEL_SMALL || size > MODEL_LARGE)
		size = MODEL_MEDIUM;
	p->size = size;

	/* Update preferences dialog */
	if (p->xml)
		prefs_update_board_size(p);
}

void
on_prefs_dialog_response(GtkWidget *widget, gint response, gpointer data)
{
	gtk_widget_destroy(widget);

	switch (response) {
	case GTK_RESPONSE_HELP:
		/* FIXME: Do help stuff */
		break;
	case GTK_RESPONSE_CLOSE:
	default:
		break;
	}
}

void
on_prefs_dialog_destroy(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	g_object_unref(G_OBJECT(p->xml));
	p->xml = NULL;
}

void
on_rb_small_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_BOARD_SIZE, MODEL_SMALL, NULL);
}

void
on_rb_normal_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_BOARD_SIZE, MODEL_MEDIUM, NULL);
}

void
on_rb_large_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_BOARD_SIZE, MODEL_LARGE, NULL);
}

void
on_rb_pvp_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_AI_LEVEL, AI_NONE, NULL);
}

void
on_rb_pveasy_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_AI_LEVEL, AI_EASY, NULL);
}

void
on_rb_pvmedium_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_AI_LEVEL, AI_MEDIUM, NULL);
}

void
on_rb_pvdiff_toggled(GtkWidget *widget, gpointer data)
{
	Prefs *p = (Prefs *)data;

	gconf_client_set_int(p->client, GCONF_KEY_AI_LEVEL, AI_DIFFICULT, NULL);
}

