/* Copyright (c) 2003 Theo van Klaveren, All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: scorebox.c,v 1.7 2005/04/06 19:32:54 klaveren Exp $
 */
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include <glib.h>
#include <gtk/gtk.h>

#include "scorebox.h"
#include "i18n.h"
#include "graphics.h"

enum {
  PROP_0,
  PROP_PLAYER_COLOR
};

struct _ScoreBoxPrivate {
	GtkWidget	*image;
	GtkWidget	*label;
	GtkWidget	*score;
};

static void		 scorebox_init(ScoreBox *sb);
static void		 scorebox_class_init(ScoreBoxClass *sbc);
static void		 scorebox_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void		 scorebox_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void		 scorebox_set_player_color(ScoreBox *sb, PlayerColor pc);

static GtkHBoxClass	*parent_class = NULL;

GType
scorebox_get_type()
{
  static GType sb_type = 0;
  
  if (!sb_type)
    {
      static const GTypeInfo sb_info =
      {
	sizeof (ScoreBoxClass),
	NULL,		/* base_init */
	NULL,		/* base_finalize */
	(GClassInitFunc) scorebox_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (ScoreBox),
	0,		/* n_preallocs */
	(GInstanceInitFunc) scorebox_init,
      };
      
      sb_type =
	g_type_register_static (GTK_TYPE_HBOX, "ScoreBox",
				&sb_info, 0);
    }
  
  return sb_type;
}

static void
scorebox_init(ScoreBox *sb)
{
	sb->priv = g_new0(ScoreBoxPrivate, 1);
	sb->priv->score = gtk_label_new("0");

	gtk_box_pack_end(GTK_BOX(sb), sb->priv->score, FALSE, TRUE, 0);
	gtk_box_set_spacing(GTK_BOX(sb), 6);
	gtk_container_set_border_width(GTK_CONTAINER(sb), 12);
}

static void
scorebox_class_init(ScoreBoxClass *sbc)
{
	GObjectClass *gobject_class;

	parent_class = g_type_class_peek_parent (sbc);
	gobject_class = (GObjectClass*) sbc;

	gobject_class->set_property = scorebox_set_property;
	gobject_class->get_property = scorebox_get_property;

	g_object_class_install_property (gobject_class,
					 PROP_PLAYER_COLOR,
					 g_param_spec_int ("playercolor",
							   "Player's color",
							   "Color of the player whose score is shown",
							   PC_BLUE, PC_RED, PC_BLUE,
							   G_PARAM_READABLE | G_PARAM_WRITABLE |
							   G_PARAM_CONSTRUCT_ONLY));
}

static void
scorebox_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	ScoreBox *sb = SCOREBOX(object);
	PlayerColor pc;
  
	switch (prop_id) {
	case PROP_PLAYER_COLOR:
		pc = g_value_get_int(value);
		scorebox_set_player_color(sb, pc);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}
}

static void
scorebox_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	ScoreBox *sb = SCOREBOX(object);
  
	switch (prop_id) {
	case PROP_PLAYER_COLOR:
		g_value_set_int(value, sb->color);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}
}

static void
scorebox_set_player_color(ScoreBox *sb, PlayerColor pc)
{
	gchar *text;
	sb->color = pc;

	if (pc == PC_RED) {
		text = g_strconcat("<span weight=\"bold\">",
					_("Red:"), "</span>", NULL);
		sb->priv->image = gtk_image_new_from_pixbuf(graphics_get_for_cell(CELL_BALL_RED));
		/* FIXME this leaks a ref to the pixbuf */
	} else {
		text = g_strconcat("<span weight=\"bold\">",
					_("Blue:"), "</span>", NULL);
		sb->priv->image = gtk_image_new_from_pixbuf(graphics_get_for_cell(CELL_BALL_BLUE));
		/* FIXME this leaks a ref to the pixbuf */
	}

	sb->priv->label = gtk_label_new(text);
	gtk_label_set_use_markup(GTK_LABEL(sb->priv->label), TRUE);
	g_free(text);
	gtk_box_pack_start(GTK_BOX(sb), sb->priv->image, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(sb), sb->priv->label, FALSE, TRUE, 0);
}

void
scorebox_set_score(ScoreBox *sb, gint score)
{
	gchar *text;

	text = g_strdup_printf("%d", score);
	gtk_label_set_text(GTK_LABEL(sb->priv->score), text);
	g_free(text);
}

GtkWidget *
scorebox_new(PlayerColor color)
{
	return g_object_new(SCOREBOX_TYPE, "playercolor", color, NULL);
}
