#include <glib.h>
#include <glib-object.h>

#include "boardmodel.h"

/* private data */
typedef struct _BoardModelPrivate {
	gint		 width;
	gint		 height;
	gint		*cells;		/* array of gint with content of board */
	GSList		*changes;	/* list of changes so far */
	gboolean	 frozen;	/* are changes frozen? */
} BoardModelPrivate;

#define BOARD_MODEL_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), BOARD_MODEL_TYPE, BoardModelPrivate))

/* function definitions */
static void	 _board_cells_alloc	(BoardModel *model);
static void	 _emit_changes		(BoardModel *model);

/* signal definitions */
enum {
	BOARD_SIZE_CHANGED_SIGNAL,	/* board size was changed */
	BOARD_CELL_CHANGED_SIGNAL,	/* one or more cells changed */
	LAST_SIGNAL
};
static guint board_model_signals[LAST_SIGNAL] = { 0, 0 };

G_DEFINE_TYPE(BoardModel, board_model, G_TYPE_OBJECT);

static void
board_model_class_init(BoardModelClass *klass)
{
	g_type_class_add_private(klass, sizeof(BoardModelPrivate));

	board_model_signals[BOARD_SIZE_CHANGED_SIGNAL] =
		g_signal_new("board-size-changed",
				G_TYPE_FROM_CLASS(klass),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET(BoardModelClass, board_size_changed),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0);
	board_model_signals[BOARD_CELL_CHANGED_SIGNAL] =
		g_signal_new("board-cell-changed",
				G_TYPE_FROM_CLASS(klass),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET(BoardModelClass, board_cell_changed),
				NULL, NULL,
				g_cclosure_marshal_VOID__BOXED,
				G_TYPE_NONE, 1, G_TYPE_POINTER);
}

static void
board_model_init(BoardModel *bm)
{
}

static void
_board_cells_alloc(BoardModel *model)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);

	if (priv->cells)
		g_free(priv->cells);
	priv->cells = g_malloc0(priv->height * priv->width * sizeof(gint));
}

static void
_emit_changes(BoardModel *model)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);

	g_signal_emit(G_OBJECT(model),
		      board_model_signals[BOARD_CELL_CHANGED_SIGNAL], 0,
		      priv->changes, NULL);
	priv->changes = NULL;
}

BoardModel *
board_model_new()
{
	GObject *obj;

	obj = g_object_new(BOARD_MODEL_TYPE, NULL);
	return BOARD_MODEL(obj);
}

void
board_model_set_dimensions(BoardModel *model, gint w, gint h)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);

	priv->width = w;
	priv->height = h;

	g_signal_emit(G_OBJECT(model),
		      board_model_signals[BOARD_SIZE_CHANGED_SIGNAL], 0);
	_board_cells_alloc(model);
}

void
board_model_get_dimensions(BoardModel *model, gint *w, gint *h)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);

	g_return_if_fail(w != NULL);
	g_return_if_fail(h != NULL);

	*w = priv->width;
	*h = priv->height;
}

void
board_model_freeze(BoardModel *model)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);
	if (priv->frozen == TRUE)
		g_warning("Freezing already frozen model\n");
	priv->frozen = TRUE;
}

void
board_model_thaw(BoardModel *model)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);
	if (priv->frozen == FALSE)
		g_warning("Thawing already thawed model\n");
	priv->frozen = FALSE;
	if (priv->changes != NULL)
		_emit_changes(model);
}

/* Set and get a single cell's contents. */
void
board_model_cell_set(BoardModel *model, gint x, gint y, gint content)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);
	BoardModelChange *ch = g_new0(BoardModelChange, 1);

	g_assert (x < priv->width);
	g_assert (y < priv->height);

	ch->x = x;
	ch->y = y;
	priv->cells[(y*priv->width)+x] = content;
	priv->changes = g_slist_prepend(priv->changes, ch);
	if (priv->frozen == FALSE)
		_emit_changes(model);
}

gint
board_model_cell_get(BoardModel *model, gint x, gint y)
{
	BoardModelPrivate *priv = BOARD_MODEL_GET_PRIVATE(model);

	g_assert(x < priv->width);
	g_assert(y < priv->height);

	return priv->cells[y*(priv->width)+x];
}

void
board_model_clone(BoardModel *orig, BoardModel *copy)
{
	BoardModelPrivate *origpriv = BOARD_MODEL_GET_PRIVATE(orig);
	BoardModelPrivate *copypriv = BOARD_MODEL_GET_PRIVATE(copy);

	/* FIXME: Copy changes as well? ATM don't copy if frozen */
	if (origpriv->frozen)
		g_error("Will not clone frozen BoardModel.\n");

	copypriv->width = origpriv->width;
	copypriv->height = origpriv->height;
	if (copypriv->cells != NULL)
		g_free(copypriv->cells);
	copypriv->cells = g_memdup(origpriv->cells, origpriv->width * origpriv->height * sizeof(gint));
}

