#include <gtk/gtk.h>

#include "boardview.h"
#include "boardmarshal.h"

#define MIN_WIDTH 5
#define MIN_HEIGHT 5

/* private data */
typedef struct _BoardViewPrivate {
	BoardModel	*model;

	gint		 width;		/* horizontal cell size in pixels */
	gint		 height;	/* vertical cell size in pixels */

	gint		 originx;	/* coordinates of the origin of the board (upperleft) */
	gint		 originy;

	gboolean	 square;	/* whether this board should be square */
} BoardViewPrivate;

#define BOARD_VIEW_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), BOARD_VIEW_TYPE, BoardViewPrivate))

/* private funcs */
static void	 _calc_cell_sizes		(BoardView *view, BoardModel *model, gint width, gint height);
static void	 _calc_cell_for_coords		(BoardView *view, gint x, gint y, gint *cellx, gint *celly);
static void	 _calc_coords_for_cell		(BoardView *view, gint cellx, gint celly, gint *x, gint *y);
static void	 _cell_paint			(BoardView *view, gint cellx, gint celly);

/* signal callbacks */
static gboolean	 _board_view_clicked_cb		(GtkWidget *widget,
						 GdkEventButton *event,
						 gpointer user_data);
static gboolean	 _board_view_configure_cb 	(GtkWidget *widget,
						 GdkEventConfigure *event,
						 gpointer user_data);
static gboolean	 _board_view_expose_cb	 	(GtkWidget *widget,
						 GdkEventExpose *event,
						 gpointer user_data);
static void	 _board_model_cell_changed_cb	(GObject *object, GSList *changes, gpointer user_data);
static void	 _board_model_size_changed_cb	(GObject *object, gpointer user_data);

/* signal definitions */
enum {
	BOARD_CLICKED_SIGNAL,		/* board was clicked */
	BOARD_RESIZED_SIGNAL,		/* board was resized */
	LAST_SIGNAL
};
static guint board_view_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(BoardView, board_view, GTK_TYPE_DRAWING_AREA);

static void
board_view_class_init(BoardViewClass *klass)
{
	g_type_class_add_private(klass, sizeof(BoardViewPrivate));

	board_view_signals[BOARD_CLICKED_SIGNAL] =
		g_signal_new("board-clicked",
				G_TYPE_FROM_CLASS(klass),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET(BoardViewClass, board_clicked),
				NULL, NULL,
				g_cclosure_user_marshal_VOID__INT_INT,
				G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
	board_view_signals[BOARD_RESIZED_SIGNAL] =
		g_signal_new("board-resized",
				G_TYPE_FROM_CLASS(klass),
				G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
				G_STRUCT_OFFSET(BoardViewClass, board_resized),
				NULL, NULL,
				g_cclosure_user_marshal_VOID__INT_INT,
				G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
}

static void
board_view_init(BoardView *bv)
{
	gtk_widget_add_events(GTK_WIDGET(bv), GDK_BUTTON_PRESS_MASK |
					 GDK_BUTTON_RELEASE_MASK);
	g_signal_connect(G_OBJECT(bv), "configure-event",
			 G_CALLBACK(_board_view_configure_cb), NULL);
	g_signal_connect(G_OBJECT(bv), "expose-event",
			 G_CALLBACK(_board_view_expose_cb), NULL);
	g_signal_connect(G_OBJECT(bv), "button-release-event",
			 G_CALLBACK(_board_view_clicked_cb), NULL);
}

void
board_view_set_model(BoardView *view, BoardModel *model)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	GtkWidget *widget = GTK_WIDGET(view);

	/* FIXME: remove previous model signals and stuff */
	if (priv->model != NULL)
		g_object_unref(priv->model);
	g_object_ref(model);
	priv->model = model;
	g_signal_connect(G_OBJECT(model), "board-cell-changed",
			 G_CALLBACK(_board_model_cell_changed_cb), view);
	g_signal_connect(G_OBJECT(model), "board-size-changed",
			 G_CALLBACK(_board_model_size_changed_cb), view);
	if (GTK_WIDGET_REALIZED(widget) && widget->allocation.width > 0
			&& widget->allocation.height > 0)
		_calc_cell_sizes(view, model, widget->allocation.width,
				 widget->allocation.height);
}

BoardModel *
board_view_get_model(BoardView *view)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);

	return priv->model;
}

void
board_view_set_square(BoardView *view, gboolean square)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);

	priv->square = square;
}

gboolean
board_view_get_square(BoardView *view)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	
	return priv->square;
}

GtkWidget *
board_view_new()
{
	GObject *obj;

	obj = g_object_new(BOARD_VIEW_TYPE, NULL);
	return GTK_WIDGET(obj);
}

void
board_view_draw_empty_cell(BoardView *view, gint x, gint y, gint width, gint height)
{
	GtkWidget *widget = GTK_WIDGET(view);

	gdk_draw_rectangle(widget->window,
			   widget->style->bg_gc[0],
			   TRUE,
			   x, y, width, height);
	gdk_draw_rectangle(widget->window,
			   widget->style->bg_gc[1],
			   FALSE,
			   x, y, width - 1, height - 1);
}

/*
 * Calculate the pixel size of each cell. Clamps the size to a minimum.
 * Sets the minimum size request of the view to (cells * minimum size) pixels.
 */
static void
_calc_cell_sizes(BoardView *view, BoardModel *model, gint width, gint height)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	gint x, y;

	board_model_get_dimensions(model, &x, &y);
	priv->width = width / x;
	priv->height = height / y;
	if (priv->width <= MIN_WIDTH)
		priv->width = MIN_WIDTH;
	if (priv->height <= MIN_HEIGHT)
		priv->height = MIN_HEIGHT;

	if (priv->square) {
		priv->width = MIN(priv->height, priv->width);
		priv->height = MIN(priv->width, priv->height);
	}
	gtk_widget_set_size_request(GTK_WIDGET(view), MIN_WIDTH * x, MIN_HEIGHT * y);

	/* center the board, calculate origin */
	priv->originx = (width - (priv->width * x)) / 2;
	priv->originy = (height - (priv->height * y)) / 2;

	g_signal_emit(G_OBJECT(view), board_view_signals[BOARD_RESIZED_SIGNAL], 0,
			priv->width, priv->height);
}

static void
_calc_cell_for_coords(BoardView *view, gint x, gint y, gint *cellx, gint *celly)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);

	g_return_if_fail(cellx != NULL);
	g_return_if_fail(celly != NULL);

	x -= priv->originx;
	y -= priv->originy;

	/* Because from -width to 0 the division is also 0, make sure
	 * the return value is negative if x is negative. Ditto for y. */
	if (x < 0)
		x = -priv->width;
	if (y < 0)
		y = -priv->height;

	*cellx = x / priv->width;
	*celly = y / priv->height;
}

static void
_calc_coords_for_cell(BoardView *view, gint cellx, gint celly, gint *x, gint *y)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	gint tx, ty;

	g_return_if_fail(x != NULL);
	g_return_if_fail(x != NULL);

	tx = cellx * priv->width;
	ty = celly * priv->height;
	*x = tx + priv->originx;
	*y = ty + priv->originy;
}

static void
_cell_paint(BoardView *view, gint cellx, gint celly)
{
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	BoardViewClass *klass = BOARD_VIEW_GET_CLASS(view);
	BoardModel *model = priv->model;
	gint boardx, boardy;
	gint realx, realy;
	gint content;

	/* Don't paint outside of the board. */
	board_model_get_dimensions(model, &boardx, &boardy);
	if (cellx > (boardx - 1))
		return;
	if (celly > (boardy - 1))
		return;
	if (cellx < 0)
		return;
	if (celly < 0)
		return;

	content = board_model_cell_get(model, cellx, celly);
	_calc_coords_for_cell(view, cellx, celly, &realx, &realy);
	if (klass->draw_cell) {
		klass->draw_cell(view, realx, realy,
				 priv->width, priv->height, content);
	} else {
		board_view_draw_empty_cell(view, realx, realy,
				priv->width, priv->height);
	}
}

static gboolean
_board_view_clicked_cb(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	BoardView *view = BOARD_VIEW(widget);
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	BoardModel *model = priv->model;
	gint x, y, boardx, boardy;

	board_model_get_dimensions(model, &boardx, &boardy);
	_calc_cell_for_coords(view, event->x, event->y, &x, &y);
	if (x > (boardx - 1))
		return FALSE;
	if (y > (boardy - 1))
		return FALSE;
	if (x < 0)
		return FALSE;
	if (y < 0)
		return FALSE;
	g_signal_emit(G_OBJECT(view), board_view_signals[BOARD_CLICKED_SIGNAL], 0,
			x, y);
	return TRUE;
}

static gboolean
_board_view_configure_cb(GtkWidget *widget, GdkEventConfigure *event, gpointer user_data)
{
	BoardView *view = BOARD_VIEW(widget);
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);

	if (priv->model != NULL)
		_calc_cell_sizes(view, priv->model, event->width, event->height);
	return TRUE;
}

static gboolean
_board_view_expose_cb(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
	BoardView *view = BOARD_VIEW(widget);
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
#if 0
	GdkRectangle rect;
#endif
	gint begin_x, begin_y, end_x, end_y;
	gint i,j;

	if (priv->model == NULL)
		return FALSE;

	gdk_window_begin_paint_rect(widget->window, &event->area);

	/* Figure out which cells were exposed. */
	_calc_cell_for_coords(view, event->area.x, event->area.y, &begin_x, &begin_y);
	_calc_cell_for_coords(view, event->area.x + event->area.width,
			            event->area.y + event->area.height,
				    &end_x, &end_y);

	/* Redraw them all in one go. */
#if 0
	_calc_coords_for_cell(view, begin_x, begin_y, &rect.x, &rect.y);
	rect.width = ((end_x - begin_x) + 1) * priv->width;
	rect.height = ((end_y - begin_y) + 1) * priv->height;
	gdk_window_begin_paint_rect(widget->window, &rect);
#endif
	for (i = begin_x; i <= end_x; i++)
		for (j = begin_y; j <= end_y; j++)
			_cell_paint(view, i, j);
	gdk_window_end_paint(widget->window);

	return TRUE;
}

static void
_board_model_cell_changed_cb(GObject *object, GSList *changes, gpointer user_data)
{
	GtkWidget *widget = GTK_WIDGET(user_data);
	BoardView *view = BOARD_VIEW(user_data);
	BoardViewPrivate *priv = BOARD_VIEW_GET_PRIVATE(view);
	BoardModelChange *ch;
	GdkRectangle rect;
	gint tx, ty, bx, by;	/* top, bottom changed cells */
	int i,j;

	if (GTK_WIDGET_REALIZED(GTK_WIDGET(user_data)) == FALSE)
		return;

	/* Figure out which area to repaint. */
	tx = ty = G_MAXINT;
	bx = by = -1;
	while (changes) {
		ch = (BoardModelChange *)changes->data;
		if (ch->x < tx)
			tx = ch->x;
		if (ch->y < ty)
			ty = ch->y;
		if (ch->x > bx)
			bx = ch->x;
		if (ch->y > by)
			by = ch->y;
		g_free(ch);
		changes = g_slist_delete_link(changes, changes);
	}

	/* Redraw all cells in the area in one go. */
	_calc_coords_for_cell(view, tx, ty, &rect.x, &rect.y);
	rect.width = ((bx - tx) + 1) * priv->width;
	rect.height = ((by - ty) + 1) * priv->height;
	gdk_window_begin_paint_rect(widget->window, &rect);
	for (i = tx; i <= bx; i++)
		for (j = ty; j <= by; j++)
			_cell_paint(view, i, j);
	gdk_window_end_paint(widget->window);
}

static void
_board_model_size_changed_cb(GObject *object, gpointer user_data)
{
	BoardView *view = BOARD_VIEW(user_data);
	GtkWidget *widget = GTK_WIDGET(user_data);

	_calc_cell_sizes(view, BOARD_MODEL(object), widget->allocation.width,
			 widget->allocation.height);
}

